# Copyright (c) 2010 Florian Friesdorf <flo@chaoflow.net>
# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

import io
import logging
import restarter
import restarter.proc
import subprocess
import time


class Service:
    """Represents a system service."""

    def __init__(self, proc, restart_cmd=None, hard_restart_cmd=None, delay=0):
        self.proc = proc
        self.restart_cmd = restart_cmd
        self.hard_restart_cmd = hard_restart_cmd
        self.delay = delay
        self.restart_pending = False
        self.log = logging.getLogger('restarter')

    def check_restart(self):
        """Determine if a restart is necessary and perform it."""
        self.restart_pending = self.proc.stale
        if self.restart_pending:
            self.restart()

    def restart(self):
        """Perform service restart if restart_cmd is defined.

        If restart_cmd does not have the desired effect and hard_restart_cmd is
        defined, take one more attempt but fail if that does not help, too.
        """

        if self.restart_cmd:
            try:
                self._restart_verify(self.restart_cmd)
            except restarter.RestartError as e:
                if self.hard_restart_cmd:
                    self.log.warn(str(e))
                    self._restart_verify(self.hard_restart_cmd)
                else:
                    raise
            self.restart_pending = False

    def _restart_verify(self, cmd):
        """Handle low-level service restart.

        A restart command is executed and the process is checked. Verify that
        the service is actually running and that the new process is different
        from the old one.
        """

        self.log.info('trying to restart using {}'.format(cmd))
        rc = subprocess.call(cmd, shell=True, stdin=open('/dev/null', 'rb'))
        if rc != 0:
            raise restarter.RestartError(
                'restart command {} failed with exit status {}'.format(
                    cmd, rc))
        time.sleep(self.delay)
        try:
            newproc = self.proc.reread()
        except restarter.PidfileError:
            raise restarter.RestartError(
                'failed, no new service pid found for {}'.format(
                    self.proc.pidfile))
        if newproc == self.proc:
            raise restarter.RestartError(
                'failed, old process {} is still running'.format(
                    self.proc.pid))



# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

import logging
import mock
import os
import restarter
import restarter.proc
import tempfile
import time
import unittest
from restarter.service import Service


class ServiceTest(unittest.TestCase):

    def setUp(self):
        # silence logger
        logging.getLogger('restarter').setLevel(logging.CRITICAL+1)

    @mock.patch('restarter.service.Service._restart_verify')
    def test_check_should_call_restart_if_stale(self, _rv):
        proc = restarter.proc.Proc('pidfile', 4711)
        proc.exe = '(deleted)'
        s = Service(proc, 'cmd1')
        s.check_restart()
        self.assertEqual(_rv.call_count, 1)

    @mock.patch('restarter.service.Service._restart_verify')
    def test_restart_should_do_nothing_if_no_restart_cmd(self, _rv):
        s = Service('dummy proc')
        s.restart()
        self.assertEqual(_rv.call_count, 0)

    @mock.patch('restarter.service.Service._restart_verify')
    def test_restart_should_call_restart_verify(self, _rv):
        s = Service('dummy proc', 'cmd1')
        s.restart()
        self.assertGreater(_rv.call_count, 0)

    @mock.patch('restarter.service.Service._restart_verify')
    def test_restart_should_try_once_unless_hard_restart_cmd(self, _rv):
        _rv.side_effect = restarter.RestartError('restart failed')
        s = Service('dummy proc', 'cmd1')
        self.assertRaises(restarter.RestartError, s.restart)
        self.assertEqual(_rv.call_count, 1)

    @mock.patch('restarter.service.Service._restart_verify')
    def test_restart_should_try_twice_if_hard_restart_cmd(self, _rv):
        _rv.side_effect = restarter.RestartError('restart failed')
        s = Service('dummy proc', 'cmd1', 'cmd2')
        self.assertRaises(restarter.RestartError, s.restart)
        self.assertEqual(_rv.call_count, 2)


class ServiceRestartVerifyTest(unittest.TestCase):

    def setUp(self):
        self.pidfile = tempfile.NamedTemporaryFile(mode='w+')
        print(os.getpid(), file=self.pidfile)
        self.pidfile.seek(0)
        self.proc = restarter.proc.Proc(self.pidfile.name, os.getpid())

    def tearDown(self):
        self.pidfile.close()

    @mock.patch('subprocess.call')
    def test_rv_should_call_command(self, _call):
        _call.return_value = 0
        s = Service(self.proc)
        try:
            s._restart_verify('cmd')
        except restarter.RestartError:
            pass
        self.assertEqual(1, _call.call_count)

    def test_rv_should_raise_on_command_failure(self):
        s = Service(self.proc)
        self.assertRaises(restarter.RestartError, s._restart_verify, 'false')

    def test_rv_should_raise_if_no_new_pidfile(self):
        s = Service(self.proc)
        with mock.patch.object(self.proc, 'reread') as _reread:
            _reread.side_effect = restarter.PidfileError()
            self.assertRaises(restarter.RestartError, s._restart_verify, 'true')

    def test_rv_should_raise_if_same_pid(self):
        s = Service(self.proc)
        self.assertRaises(restarter.RestartError, s._restart_verify, 'true')

    @mock.patch('time.sleep')
    def test_rv_should_sleep(self, _sleep):
        s = Service(self.proc, 'true', delay=3)
        try:
            s.restart()
        except restarter.RestartError:
            pass
        _sleep.assert_called_with(3)




from setuptools import setup, find_packages
import sys, os, os.path

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()
CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()
version = open(os.path.join(here, 'version.txt')).read().strip()


setup(name='restarter',
    version=version,
    description="Automatic service restart after updates",
    long_description=README + '\n\n' + CHANGES,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Zope Public License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.1',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
    author='Christian Kauhaus',
    author_email='kc@gocept.com',
    url='http://code.gocept.com/hg/public/restarter',
    license='ZPL',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts':
            ['restarter=restarter:main']
    },
    tests_require=['mock'],
    test_suite='restarter'
)



# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

import contextlib
import mock
import os
import restarter
import sys
import tempfile
import unittest
from restarter.proc import Proc, readpid


class ReadpidTest(unittest.TestCase):

    @contextlib.contextmanager
    def self_in_pidfile(self):
        with tempfile.NamedTemporaryFile(mode='w+') as f:
            print(os.getpid(), file=f)
            f.seek(0)
            yield f.name

    def test_should_contain_exe(self):
        with self.self_in_pidfile() as pidfile:
            proc = readpid(pidfile)
            self.assertEqual(os.path.realpath(sys.executable), proc.exe)

    def test_should_contain_maps(self):
        with self.self_in_pidfile() as pidfile:
            proc = readpid(pidfile)
            self.assertIn('/libc-', proc.maps)

    def test_should_fail_unless_pidfile(self):
        tf = tempfile.NamedTemporaryFile()
        filename = tf.name
        tf.close() # unlinks file
        self.assertRaises(restarter.PidfileError, readpid, filename)

    @mock.patch('os.readlink')
    def test_should_require_running_process(self, _readlink):
        _readlink.side_effect = OSError(2, 'No such file or directory')
        with self.self_in_pidfile() as fn:
            self.assertRaises(restarter.PidfileError, readpid, fn)

    def test_should_require_integer_pid(self):
        with tempfile.NamedTemporaryFile(mode='w+') as f:
            self.assertRaises(restarter.PidfileError, readpid, f.name)

    def test_reread_should_return_new_proc(self):
        with self.self_in_pidfile() as pidfile:
            proc = readpid(pidfile)
            proc2 = proc.reread()
            self.assertEqual(proc, proc2)


class ProcTest(unittest.TestCase):

    atd_maps_intact = """\
08048000-0804c000 r-xp 00000000 fd:01 853944     /usr/sbin/atd
08ecc000-08eed000 rw-p 00000000 00:00 0          [heap]
b75e0000-b75ea000 r-xp 00000000 fd:01 622557     /lib/libnss_files-2.11.2.so
b75ec000-b75ee000 rw-p 00000000 00:00 0 
b7738000-b7743000 r-xp 00000000 fd:01 622603     /lib/libpam.so.0.83.0
b774e000-b7750000 rw-p 00000000 00:00 0 
b7750000-b7751000 r-xp 00000000 00:00 0          [vdso]
b7751000-b776d000 r-xp 00000000 fd:01 622562     /lib/ld-2.11.2.so
bfd0f000-bfd30000 rw-p 00000000 00:00 0          [stack]
"""

    atd_maps_deleted = """\
08048000-0804c000 r-xp 00000000 fd:01 853944     /usr/sbin/atd
08ecc000-08eed000 rw-p 00000000 00:00 0          [heap]
b75e0000-b75ea000 r-xp 00000000 fd:01 622557     /lib/libnss_files-2.11.2.so
b75ec000-b75ee000 rw-p 00000000 00:00 0 
b7738000-b7743000 r-xp 00000000 fd:01 622603     /lib/libpam.so.0.83.0 (deleted)
b774e000-b7750000 rw-p 00000000 00:00 0 
b7750000-b7751000 r-xp 00000000 00:00 0          [vdso]
b7751000-b776d000 r-xp 00000000 fd:01 622562     /lib/ld-2.11.2.so
bfd0f000-bfd30000 rw-p 00000000 00:00 0          [stack]
"""

    @mock.patch('os.access')
    def test_intact_exe_should_not_be_stale(self, _access):
        _access.return_value = True
        p = Proc('pidfile', 0)
        p.exe = '/path/to/exe'
        self.assertFalse(p.stale_exe())

    @mock.patch('os.access')
    def test_deleted_exe_should_be_stale(self, _access):
        _access.return_value = False
        p = Proc('pidfile', 0)
        p.exe = '/path/to/exe (deleted)'
        self.assertTrue(p.stale_exe())

    @mock.patch('os.access', new=lambda fn, mode: not '(deleted)' in fn)
    def test_intact_maps_should_not_be_stale(self):
        p = Proc('pidfile', 0)
        p.maps = self.atd_maps_intact
        self.assertFalse(p.stale_maps())

    @mock.patch('os.access', new=lambda fn, mode: not '(deleted)' in fn)
    def test_deleted_maps_should_be_stale(self):
        p = Proc('pidfile', 0)
        p.maps = self.atd_maps_deleted
        self.assertTrue(p.stale_maps())

    @mock.patch('os.access', new=lambda fn, mode: not '(deleted)' in fn)
    def test_exclude_maps_using_filename(self):
        p = Proc('pidfile', 0, exclude=['/irrelevant', '/lib/libpam.so.0.83.0',
                                        '/dummy'])
        p.maps = self.atd_maps_deleted
        self.assertFalse(p.stale_maps())

    @mock.patch('os.access', new=lambda fn, mode: not '(deleted)' in fn)
    def test_exclude_maps_using_glob(self):
        p = Proc('pidfile', 0, exclude=['*pam.so*'])
        p.maps = self.atd_maps_deleted
        self.assertFalse(p.stale_maps())



# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

"""restarter command line interface."""


import logging
import optparse
import restarter.app
import pkg_resources
import sys


def options():
    dist = pkg_resources.get_distribution('restarter')
    p = optparse.OptionParser(
        version='%prog '+dist.version,
        usage='%prog [OPTIONS] PIDFILE')
    p.description = 'Perform service restarts after updates'
    p.add_option('-r', '--restart', metavar='COMMAND',
                 help='shell command to restart service')
    p.add_option('-R', '--hard-restart', metavar='COMMAND',
                 help='shell command to perform hard restart if normal restart '
                 'has failed')
    p.add_option('-x', '--exclude', action='append', metavar='PATTERN',
                 help="ignore mmap'ed files mathing shell glob PATTERN in the "
                 "decision to restart")
    p.add_option('-d', '--delay', type='float', default=0,
                 help='wait DELAY seconds between restart command execution '
                 'and service startup verification')
    p.add_option('-f', '--fail', action='store_true',
                 help='fail if pidfile does not exist or process is not '
                 'running')
    p.add_option('-v', '--verbose', action='store_const', const=1,
                 dest='verbosity', default=0,
                 help='print more output')
    p.add_option('-q', '--quiet', action='store_const', const=-1,
                 dest='verbosity', help='print only error messages')
    options, args = p.parse_args()
    if len(args) != 1:
        p.error('need pid file')
    return args[0], options


def setup_logger(verbosity=0):
    l = logging.getLogger('restarter')
    ch = logging.StreamHandler()
    if verbosity < 0:
        ch.setLevel(logging.ERROR)
    elif verbosity > 0:
        ch.setLevel(logging.INFO)
    else:
        ch.setLevel(logging.WARNING)
    ch.setFormatter(logging.Formatter('%(name)s: %(message)s'))
    l.addHandler(ch)


def main():
    pidfile, opts = options()
    setup_logger(opts.verbosity)
    app = restarter.app.App(pidfile, opts)
    try:
        restart_pending = app()
    except RestartError as e:
        print('restarter: ' + str(e), file=sys.stderr)
        sys.exit(1)
    if restart_pending:
        sys.exit(3)



# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

from restarter.cli import main


# common exceptions

class PidfileError(EnvironmentError):
    pass


class RestartError(RuntimeError):
    pass



# Copyright (c) 2010 Florian Friesdorf <flo@chaoflow.net>
# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

import errno
import fnmatch
import logging
import os
import restarter


class Proc:
    """Represent a process identified by a pidfile.

    Processes may be running or not running. If running, we can examine them for
    stale executables or memory mapped objects.
    """

    def __init__(self, pidfile, pid, exclude=None):
        self.pidfile = pidfile
        self.pid = pid
        self.exclude = exclude
        self.exe = None
        self.maps = None
        self.log = logging.getLogger('restarter')

    def __eq__(self, other):
        return self.pidfile == other.pidfile and self.pid == other.pid

    @property
    def stale(self):
        """Determine if this process has any stale objects."""
        return self.stale_exe() or self.stale_maps()

    def stale_exe(self):
        """Determine if the executable is stale."""
        if not os.access(self.exe, os.F_OK):
            self.log.info('stale exe {}'.format(self.exe))
            return True
        return False

    def stale_maps(self):
        """Determine if there are any stale (non-excluded) mmap'ed objects."""
        for line in self.maps.split('\n'):
            try:
                filename = line.split(None, 5)[5]
            except IndexError:
                continue
            if not filename.startswith('/'):
                continue
            if (not self.match_exclude(filename) and
                not os.access(filename, os.F_OK)):
                self.log.info('stale mmap {}'.format(filename))
                return True
        return False

    def match_exclude(self, filename):
        """Filter object list for exceptions.

        Exclude expressions may be specified as prefixes (e.g., /lib/libc) or
        shell glob expressions. Please note that the kernel tends to add some
        informative string to stale objects (e.g., "(deleted)"), so be sure to
        leave the suffix variable.
        """

        if self.exclude:
            if any(filename.startswith(e) for e in self.exclude):
                return True
            if any(fnmatch.fnmatch(filename, e) for e in self.exclude):
                return True
        return False

    def reread(self):
        """Reread pidfile and return new Proc object with up-to-date status."""
        return readpid(self.pidfile)


def readpid(pidfile, exclude=None):
    """Factory to create Proc objects from pidfiles."""
    log = logging.getLogger('restarter')
    try:
        pid = int(open(pidfile).read().strip())
        proc = Proc(pidfile, pid, exclude)
        proc.exe = os.readlink('/proc/{}/exe'.format(pid)).strip()
        proc.maps = open('/proc/{}/maps'.format(pid)).read()
        return proc
    except EnvironmentError as e:
        log.info('service not running: cannot open {}'.format(e.filename))
        raise restarter.PidfileError(e.errno, e.strerror, e.filename)
    except ValueError as e:
        log.info('invalid pidfile')
        raise restarter.PidfileError(errno.EINVAL, str(e), pidfile)



# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

import logging
import os
import restarter
import restarter.service
import restarter.proc


class App:
    """Main application skeleton."""

    def __init__(self, pidfile, options):
        """Set up program parameters."""
        self.pidfile = pidfile
        self.restart_cmd = options.restart
        self.hard_restart_cmd = options.hard_restart
        self.delay = options.delay
        self.fail = options.fail
        self.exclude = options.exclude
        self.perform_restart = False
        self.log = logging.getLogger('restarter')
        self.log.setLevel(logging.DEBUG)

    def __call__(self):
        """Perform program run."""
        try:
            proc = restarter.proc.readpid(self.pidfile, self.exclude)
        except restarter.PidfileError as e:
            if self.fail:
                raise restarter.RestartError(str(e))
            return
        svc = restarter.service.Service(
            proc, self.restart_cmd, self.hard_restart_cmd, self.delay)
        svc.check_restart()
        return svc.restart_pending
